import java.io.IOException;
import java.io.PrintWriter;
import java.util.ArrayList;
import java.util.Arrays;
import java.util.HashMap;

public class Coherence {
    private static final String WORD_GLOVE_FILE = "gloveMap.tsv";


    public static void genTopItemsFromMultimodalPhi(String workDir, int L, int nSelect) throws IOException{
        for(int l=0;l<L;l++){
            ArrayList<String> dict = TextUtility.readTextFileToList(workDir+"/dict-"+l+".txt");
            String phiFile = workDir+"phi-"+l+".csv";
            ArrayList<String> lines = TextUtility.readTextFileToList(phiFile);
            int M = lines.size();
            int K = lines.get(0).split(",").length;

            String[][] table = new String[M][];
            for(int i=0;i<M;i++){
                table[i] = lines.get(i).split(",");
            }
            PrintWriter pw = new PrintWriter(workDir+"topics_dict"+l+".txt");
            for(int i=0;i<K;i++){
                String[][] nameToValue = new String[M][2];
                for(int j=0;j<M;j++){
                    nameToValue[j][0] = dict.get(j);
                    nameToValue[j][1] = table[j][i];
                }
                String[] dict1List = TextUtility.sortListByDoubleKey(nameToValue, true);
                String[] topDict1 = Arrays.copyOfRange(dict1List, 0, nSelect);

                pw.println(String.join(", ", topDict1));
            }
            pw.close();
        }
    }
    public static double[] nnpsVenueAssociation(String workDir, String nnpsTopicsFile, String venueTopicsFile) {
        double[] res = new double[3];

        HashMap<String, String> wordMap = TextUtility.read2ColTable(WORD_GLOVE_FILE);
        ArrayList<String> nnpsTopics = TextUtility.readTextFileToList(workDir+nnpsTopicsFile);
        ArrayList<String> venueTopics = TextUtility.readTextFileToList(workDir+venueTopicsFile);
        int k = nnpsTopics.size();

        double nnpsCoherence[] = new double[k];
        for (int l = 0; l < k; l++) {
            double sum = 0;
            int count = 0;
            String[] tks = nnpsTopics.get(l).split(", ");
            int n = tks.length;
            for (int i = 0; i < n - 1; i++) {
                String word1 = tks[i].toLowerCase();
                if (!wordMap.containsKey(word1)) continue;
                for (int j = i + 1; j < n; j++) {
                    String word2 = tks[j].toLowerCase();
                    if (!wordMap.containsKey(word2)) continue;
                    double value = VectorUtility.cosineSimilarity(wordMap.get(word1), wordMap.get(word2));
                    sum += value;
                    count++;
                }
            }
            nnpsCoherence[l] = sum / count;
        }

        double venueCoherence[] = new double[k];
        for (int l = 0; l < k; l++) {
            double sum = 0;
            int count = 0;
            String[] tks = venueTopics.get(l).split(", ");
            int n = tks.length;
            for (int i = 0; i < n - 1; i++) {
                String[] tks1 = tks[i].split(" ");
                String word1 = tks1[tks1.length - 1].replaceAll("\\)", "").toLowerCase();
                if (!wordMap.containsKey(word1)) continue;
                for (int j = i + 1; j < n; j++) {
                    String[] tks2 = tks[j].split(" ");
                    String word2 = tks2[tks2.length - 1].replaceAll("\\)", "").toLowerCase();
                    if (!wordMap.containsKey(word2)) continue;
                    double value = VectorUtility.cosineSimilarity(wordMap.get(word1), wordMap.get(word2));
                    sum += value;
                    count++;
                }
            }
            venueCoherence[l] = sum / count;
        }

        double association[] = new double[k];
        for (int l = 0; l < k; l++) {
            String[] nnpsTks = nnpsTopics.get(l).split(", ");
            String[] venueTks = venueTopics.get(l).split(", ");
            int n = nnpsTks.length;
            if(n != venueTks.length){
                System.err.println("nnps topics has different number than venue topics!");
                System.exit(1);
            }
            double sum = 0;
            int count = 0;
            for (int i = 0; i < n; i++) {
                String word1 = nnpsTks[i].toLowerCase();
                if(! wordMap.containsKey(word1)) continue;
                count++;
                double max = 0;
                String maxWord = "";
                for (int j = 0; j < n; j++) {
                    String[] tks = venueTks[j].split(" ");
                    String word2 = tks[tks.length - 1].replaceAll("\\)", "").toLowerCase();
                    if(! wordMap.containsKey(word2)) continue;
                    double value = VectorUtility.cosineSimilarity(wordMap.get(word1), wordMap.get(word2));
                    if(max < value) {
                        max = value;
                        maxWord = venueTks[j];
                    }
                }
                sum += max;
            }
            association[l] = sum/count;
        }
        res[0] = VectorUtility.mean(nnpsCoherence);
        res[1] = VectorUtility.mean(venueCoherence);
        res[2] = VectorUtility.mean(association);
        return res;
    }

    public static double[] calCoherence(String resultDir) throws IOException {
        genTopItemsFromMultimodalPhi(resultDir, 3, 10);
        double[] res = nnpsVenueAssociation(resultDir, "topics_dict0.txt", "topics_dict1.txt");
        return res;
    }

    public static void main(String[] args) throws IOException{
        double[] res = calCoherence("./");
        System.out.println("Tweet coherence: "+res[0]);
        System.out.println("Venue coherence: "+res[1]);
        System.out.println("Tweet venue association: "+res[2]);
    }
}
