import java.io.BufferedReader;
import java.io.FileReader;
import java.io.IOException;
import java.io.PrintWriter;
import java.util.ArrayList;
import java.util.Arrays;

public class WMLDA{

    int[][][] documents;
    int[] V;
    int L;
    double[] lambda;
    int K;
    double alpha;
    double[] beta;
    int z[][][];
    int[][][] nw;
    int[][][] nd;
    int[][] nwsum;
    int[][] ndsum;
    double[][] thetasum;
    double[][][] phisum;
    int numstats;
    private static int THIN_INTERVAL = 20;
    private static int BURN_IN = 100;
    private static int ITERATIONS = 1000;
    private static int SAMPLE_LAG = 10;

    public WMLDA(int[][][] documents, int[] V, double[] lambda) {

        this.documents = documents;
        this.V = V;
        this.L = V.length;
        this.lambda = lambda;

        int wordCount = 0;
        for(int m=0;m<documents.length;m++){
            for(int l=0;l<L;l++){
                wordCount += documents[m][l].length;
            }
        }
        System.out.println("corpus total words:"+wordCount);
    }

    public void initialState(int K) {
        int M = documents.length;

        nw = new int[L][][];
        for(int l=0;l<L;l++){
            nw[l] = new int[V[l]][K];
        }
        nd = new int[M][L][K];
        nwsum = new int[L][K];
        ndsum = new int[M][L];

        z = new int[M][L][];
        for(int l=0;l<L;l++){
            for (int m = 0; m < M; m++) {
                int N = documents[m][l].length;
                z[m][l] = new int[N];
                for (int n = 0; n < N; n++) {
                    int topic = (int) (Math.random() * K);
                    z[m][l][n] = topic;
                    // number of instances of word i assigned to topic j
                    nw[l][documents[m][l][n]][topic]++;
                    // number of words in document i modal l assigned to topic j.
                    nd[m][l][topic]++;
                    // total number of words assigned to topic j.
                    nwsum[l][topic]++;
                }
                // total number of words in document i modal l
                ndsum[m][l] = N;
            }
        }
    }

    public void gibbs(int K, double alpha, double[] beta) {
        this.K = K;
        this.alpha = alpha;
        this.beta = beta;

        if (SAMPLE_LAG > 0) {
            thetasum = new double[documents.length][K];
            phisum = new double[L][K][];
            for(int l=0;l<L;l++){
                phisum[l] = new double[K][V[l]];
            }
            numstats = 0;
        }

        initialState(K);

        System.out.println("Sampling " + ITERATIONS
                + " iterations with burn-in of " + BURN_IN + " (B/S="
                + THIN_INTERVAL + ").");

        for (int i = 0; i < ITERATIONS; i++) {

            // for all z_i
            for (int m = 0; m < z.length; m++) {
                for(int l=0;l<L;l++){
                    double[][] plk = new double[L][K];
                    for(int l2=0;l2<L;l2++){
                        for(int k=0;k<K;k++){
                            plk[l2][k] = lambda[l2]  * (nd[m][l2][k] + alpha) / (ndsum[m][l2] + K * alpha);
                        }
                    }
                    for (int n = 0; n < z[m][l].length; n++) {
                        int topic = sampleFullConditional(m, l, n, plk);
                        z[m][l][n] = topic;
                    }
                }
            }

            // get statistics after burn-in
            if ((i <= BURN_IN) && (i % THIN_INTERVAL == 0)) {
                System.out.print("B");
            }
            if ((i > BURN_IN) && (i % THIN_INTERVAL == 0)) {
                updateParams();
                System.out.print("S");
            }

        }
        System.out.println();
    }

    private int sampleFullConditional(int m, int l, int n, double[][] plk) {

        // remove z_i from the count variables
        int topic = z[m][l][n];
        nw[l][documents[m][l][n]][topic]--;
        nd[m][l][topic]--;
        nwsum[l][topic]--;
        ndsum[m][l]--;

        double[] p = new double[K];
        for (int k = 0; k < K; k++) {
            int word = documents[m][l][n];
            double a = lambda[l] * (nd[m][l][k] + alpha) / (ndsum[m][l] + K * alpha);
            for(int l2=0;l2<L;l2++){
                if(l2 == l) continue;
                a += plk[l2][k];
            }
            double b = (nw[l][word][k] + beta[l]) / (nwsum[l][k] + V[l] * beta[l]);
            p[k] = a * b;
        }
        for (int k = 1; k < p.length; k++) {
            p[k] += p[k - 1];
        }
        double u = Math.random() * p[K - 1];
        for (topic = 0; topic < p.length; topic++) {
            if (u < p[topic])
                break;
        }

        // add newly estimated z_i to count variables
        nw[l][documents[m][l][n]][topic]++;
        nd[m][l][topic]++;
        nwsum[l][topic]++;
        ndsum[m][l]++;

        return topic;
    }

    private void updateParams() {
        for(int l=0;l<L;l++){
            for (int m = 0; m < documents.length; m++) {
                for (int k = 0; k < K; k++) {
                    thetasum[m][k] += lambda[l]*((nd[m][l][k] + alpha) / (ndsum[m][l] + K * alpha));
                }
            }
        }
        for(int l=0;l<L;l++){
            for (int k = 0; k < K; k++) {
                for (int w = 0; w < V[l]; w++) {
                    phisum[l][k][w] += (nw[l][w][k] + beta[l]) / (nwsum[l][k] + V[l] * beta[l]);
                }
            }
        }
        numstats++;
    }

    public double[][] getTheta() {
        double[][] theta = new double[documents.length][K];

        if (SAMPLE_LAG > 0) {
            for (int m = 0; m < documents.length; m++) {
                for (int k = 0; k < K; k++) {
                    theta[m][k] = thetasum[m][k] / numstats;
                }
            }

        } else {
            for (int m = 0; m < documents.length; m++) {
                for (int k = 0; k < K; k++) {
                    for(int l=0;l<L;l++){
                        theta[m][k] += lambda[l]*((nd[m][l][k] + alpha) / (ndsum[m][l] + K * alpha));
                    }
                }
            }
        }

        return theta;
    }

    public double[][][] getPhi() {
        double[][][] phi = new double[L][][];
        for(int l=0;l<L;l++) {
            phi[l] = new double[K][V[l]];
        }
            if (SAMPLE_LAG > 0) {
            for(int l=0;l<L;l++){
                for (int k = 0; k < K; k++) {
                    for (int w = 0; w < V[l]; w++) {
                        phi[l][k][w] = phisum[l][k][w] / numstats;
                    }
                }
            }
        } else {
            for(int l=0;l<L;l++) {
                for (int k = 0; k < K; k++) {
                    for (int w = 0; w < V[l]; w++) {
                        phi[l][k][w] = (nw[l][w][k] + beta[l]) / (nwsum[l][k] + V[l] * beta[l]);
                    }
                }
            }
        }
        return phi;
    }

    public void configure(int iterations, int burnIn, int thinInterval,
                          int sampleLag) {
        ITERATIONS = iterations;
        BURN_IN = burnIn;
        THIN_INTERVAL = thinInterval;
        SAMPLE_LAG = sampleLag;
    }

    public static int[][][] loadMultimodalIndexCorpusStrings(ArrayList<String> corpusStrings){
        int M =corpusStrings.size();
        int L = corpusStrings.get(0).split("\t").length;
        int[][][] res = new int[M][L][];
        for(int m=0;m<M;m++){
            String[] tks1 = corpusStrings.get(m).split("\t");
            for(int l=0;l<L;l++){
                String[] tks2 = tks1[l].split(" ");
                int N = tks2.length;
                int[] words = new int[N];
                for(int j=0;j<N;j++){
                    words[j] = Integer.valueOf(tks2[j]);
                }
                res[m][l] = words;
            }
        }
        return res;
    }

    public static void runCorpus(String workDir, ArrayList<String> corpusStrings, ArrayList<ArrayList<String>> dicts, double[] lambda, double[] beta, int K, int iter, int burninIter, int thinInterval) throws IOException {
        int[][][] corpus = loadMultimodalIndexCorpusStrings(corpusStrings);
        int M = corpus.length;
        int L = dicts.size();
        int[] V = new int[L];
        for(int l=0;l<L;l++){
            V[l] = dicts.get(l).size();
        }

        double alpha = 0.5;

        System.out.println("WMLDA using Gibbs Sampling.");

        WMLDA lda = new WMLDA(corpus, V, lambda);
        lda.configure(iter, burninIter, thinInterval, 10);
        lda.gibbs(K, alpha, beta);

        double[][] theta = lda.getTheta();
        double[][][] phi = lda.getPhi();

        PrintWriter pw = new PrintWriter(workDir+"theta-LDA.csv");
        for(int i=0;i<M;i++){
            ArrayList<String> temp = new ArrayList<>();
            for(int k=0;k<K;k++){
                temp.add(""+theta[i][k]);
            }
            pw.println(String.join(",", temp));
        }
        pw.close();

        for(int l=0;l<L;l++){
            pw = new PrintWriter(workDir+"phi-"+l+".csv");
            for(int n=0;n<V[l];n++){
                ArrayList<String> temp = new ArrayList<>();
                for(int k=0;k<K;k++){
                    temp.add(""+phi[l][k][n]);
                }
                pw.println(String.join(",", temp));
            }
            pw.close();
        }
    }

    public static void runCorpus(String workDir, String corpusFile, ArrayList<String> dictFiles, double[] lambda, double[] beta,
                                 int K, int iter, int burninIter, int thinInterval) throws IOException{
        ArrayList<String> lines = TextUtility.readTextFileToList(corpusFile);
        ArrayList<ArrayList<String>> dicts = new ArrayList<>();
        for(String dictFile : dictFiles){
            dicts.add(TextUtility.readTextFileToList(dictFile));
        }
        runCorpus(workDir, lines, dicts, lambda, beta, K, iter, burninIter, thinInterval);
    }
    public static void main(String[] args) throws IOException{
        if(args.length != 1){
            System.err.println("java WMLDA <configuration file>");
            System.exit(1);
        }
        ArrayList<String> lines = TextUtility.readTextFileToList(args[0]);
        String workDir = lines.get(0).split("=")[1];
        String corpusFile = lines.get(1).split("=")[1];
        ArrayList<String> dictFiles = new ArrayList<>(Arrays.asList(lines.get(2).split("=")[1].split(",")));
        double[] lambda = TextUtility.stringArrayToDoubleArray(lines.get(3).split("=")[1].split(","));
        double[] beta = TextUtility.stringArrayToDoubleArray(lines.get(4).split("=")[1].split(","));
        int K = Integer.valueOf(lines.get(5).split("=")[1]);
        int iter = Integer.valueOf(lines.get(6).split("=")[1]);
        int burnIter = Integer.valueOf(lines.get(7).split("=")[1]);
        int thinInterval = Integer.valueOf(lines.get(8).split("=")[1]);

        runCorpus(workDir, corpusFile, dictFiles, lambda, beta, K, iter, burnIter, thinInterval);
    }
}